import pathlib
import typing
from unittest import mock

import pytest

from trycicle.cache import copy_matching, get_cache_path, pull_cache, push_cache
from trycicle.models import Cache
from trycicle.parser import parse_config
from trycicle.variables import Variables, constants


def test_cache_path_simple_key() -> None:
    cache = Cache(key="test")
    variables = Variables({})
    assert (
        get_cache_path(cache, variables, pathlib.Path("project"))
        == "cache-project-test"
    )


def test_cache_path_key_variable() -> None:
    cache = Cache(key="cache-$CI_COMMIT_REF_SLUG")
    variables = Variables({"CI_COMMIT_REF_SLUG": "main"})
    assert (
        get_cache_path(cache, variables, pathlib.Path("project"))
        == "cache-project-cache-main"
    )


def test_cache_path_key_file() -> None:
    cache = Cache(key_files=["a.txt"])
    variables = Variables({})

    with mock.patch("trycicle.cache.get_command_output") as git_log:
        git_log.return_value = "sha"
        assert (
            get_cache_path(cache, variables, pathlib.Path("project"))
            == "cache-project-sha"
        )

    git_log.assert_called_once_with(
        ["git", "log", "--pretty=format:%H", "-1", "--", "a.txt"],
        cwd=pathlib.Path("project"),
    )


def test_cache_path_key_file_prefix() -> None:
    cache = Cache(key_files=["a.txt"], key_prefix="prefix")
    variables = Variables({})

    with mock.patch("trycicle.cache.get_command_output") as git_log:
        git_log.return_value = "sha"
        assert (
            get_cache_path(cache, variables, pathlib.Path("project"))
            == "cache-project-prefix-sha"
        )


def test_cache_path_key_files() -> None:
    cache = Cache(key_files=["a.txt", "b.txt"])
    variables = Variables({})

    with mock.patch("trycicle.cache.get_command_output") as git_log:
        git_log.side_effect = ["sha1", "sha2"]
        assert (
            get_cache_path(cache, variables, pathlib.Path("project"))
            == "cache-project-sha1-sha2"
        )

    assert git_log.call_count == 2


def test_recursive_matching() -> None:
    current_file = pathlib.Path(__file__)
    tests_dir = current_file.parent
    source_dir = tests_dir.parent
    variables = Variables({})
    with mock.patch("shutil.copy") as copy:
        copy_matching(["**/*.py"], variables, source_dir, pathlib.Path("destination"))

    assert copy.call_count >= 1
    copy.assert_any_call(
        current_file,
        pathlib.Path("destination") / tests_dir.name / current_file.name,
        follow_symlinks=False,
    )


def test_copy_directory() -> None:
    current_file = pathlib.Path(__file__)
    tests_dir = current_file.parent
    source_dir = tests_dir.parent
    variables = Variables({})
    with mock.patch("shutil.copytree") as copytree:
        copy_matching(
            [tests_dir.name], variables, source_dir, pathlib.Path("destination")
        )

    copytree.assert_called_once_with(
        tests_dir,
        pathlib.Path("destination") / tests_dir.name,
        symlinks=True,
        dirs_exist_ok=True,
    )


def test_copy_ignores_special(tmp_path: pathlib.Path) -> None:
    # Create a (dangling) symlink
    link = tmp_path / "link"
    link.symlink_to(tmp_path / "target")

    # Make sure the pattern matches the link
    assert link in list(tmp_path.glob("*"))

    variables = Variables({})
    with mock.patch("shutil.copy") as copy, mock.patch("shutil.copytree") as copytree:
        copy_matching(["*"], variables, tmp_path, pathlib.Path("destination"))

    copy.assert_not_called()
    copytree.assert_not_called()


@pytest.mark.parametrize("policy", ["pull", "push", "pull-push"])
@pytest.mark.parametrize("direction", ["pull", "push"])
def test_push_pull(
    policy: typing.Literal["pull", "push", "pull-push"],
    direction: typing.Literal["pull", "push"],
) -> None:
    cache = Cache(paths=["**"], policy=policy)
    cache_path = pathlib.Path(__file__).parent
    source_dir = pathlib.Path("/tmp/source")
    with mock.patch("trycicle.cache.copy_matching") as copy_matching:
        f = pull_cache if direction == "pull" else push_cache
        f(cache, Variables({}), cache_path, source_dir)
    assert copy_matching.call_count == int(direction in policy)


def test_cache_path_absolute(tmp_path: pathlib.Path) -> None:
    with open("tests/examples/cache.yml") as fp:
        config = parse_config(fp)

    job = config.get_job("cache-path-absolute")
    variables = Variables(constants) + job.variables

    assert job.cache
    assert job.cache[0].paths
    assert job.cache[0].paths[0] == "${UV_CACHE_DIR}"
    assert variables.replace(job.cache[0].paths[0]) == "/src/.cache/uv"

    with mock.patch("trycicle.cache.glob") as glob:
        pull_cache(job.cache[0], variables, tmp_path, tmp_path)

    glob.assert_any_call(tmp_path, ".cache/uv")
    glob.assert_any_call(tmp_path, ".venv")


@pytest.mark.parametrize(
    "name",
    ["cache-path-outside-project", "cache-path-relative-outside-project"],
)
def test_cache_path_outside_project(name: str, tmp_path: pathlib.Path) -> None:
    with open("tests/examples/cache.yml") as fp:
        config = parse_config(fp)

    job = config.get_job(name)
    variables = Variables(constants) + job.variables

    with mock.patch("shutil.copy") as copy, mock.patch("shutil.copytree") as copytree:
        pull_cache(job.cache[0], variables, tmp_path, tmp_path)

    copy.assert_not_called()
    copytree.assert_not_called()
